@@ -1,10 +1,9 @@
 /*  GNU ddrescue - Data recovery tool
-    Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012,
-    2013, 2014 Antonio Diaz Diaz.
+    Copyright (C) 2004-2014 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
+    the Free Software Foundation, either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
@@ -25,6 +24,7 @@
 #define _FILE_OFFSET_BITS 64
 
 #include <algorithm>
+#include <cctype>
 #include <cerrno>
 #include <climits>
 #include <cstdio>
@@ -42,6 +42,7 @@
 #include "rational.h"
 #include "block.h"
 #include "ddrescue.h"
+#include "linux.h"
 #include "loggers.h"
 
 #ifndef O_BINARY
@@ -76,7 +77,7 @@ void show_help( const int cluster, const int hardbs, const int skipbs )
                "  -h, --help                     display this help and exit\n"
                "  -V, --version                  output version information and exit\n"
                "  -a, --min-read-rate=<bytes>    minimum read rate of good areas in bytes/s\n"
-               "  -A, --try-again                mark non-split, non-trimmed blocks as non-tried\n"
+               "  -A, --try-again                mark non-trimmed, non-scraped as non-tried\n"
                "  -b, --sector-size=<bytes>      sector size of input device [default %d]\n", hardbs );
   std::printf( "  -B, --binary-prefixes          show binary multipliers in numbers [SI]\n"
                "  -c, --cluster-size=<sectors>   sectors to copy at a time [%d]\n", cluster );
@@ -89,31 +90,36 @@ void show_help( const int cluster, const int hardbs, const int skipbs )
                "  -F, --fill-mode=<types>        fill given type blocks with infile data (?*/-+)\n"
                "  -G, --generate-mode            generate approximate logfile from partial copy\n"
                "  -H, --test-mode=<file>         set map of good/bad blocks from given logile\n"
-               "  -i, --input-position=<bytes>   starting position in input file [0]\n"
+               "  -i, --input-position=<bytes>   starting position of domain in input file [0]\n"
                "  -I, --verify-input-size        verify input file size with size in logfile\n"
                "  -K, --skip-size=<min>[,<max>]  initial size to skip on read error [%sB]\n",
                format_num( skipbs, 9999, -1 ) );
-  std::printf( "  -l, --logfile-size=<entries>   do not grow logfile beyond this size [10000]\n"
-               "  -L, --loose-domain             accept an incomplete domain logfile\n"
+  std::printf( "  -L, --loose-domain             accept an incomplete domain logfile\n"
                "  -m, --domain-logfile=<file>    restrict domain to finished blocks in file\n"
                "  -M, --retrim                   mark all failed blocks as non-trimmed\n"
-               "  -n, --no-split                 skip the splitting phase\n"
+               "  -n, --no-scrape                skip the scraping phase\n"
                "  -N, --no-trim                  skip the trimming phase\n"
                "  -o, --output-position=<bytes>  starting position in output file [ipos]\n"
                "  -O, --reopen-on-error          reopen input file after every read error\n"
                "  -p, --preallocate              preallocate space on disc for output file\n"
+               "  -P, --data-preview[=<lines>]   show some lines of the latest data read [3]\n"
                "  -q, --quiet                    suppress all messages\n"
                "  -r, --retry-passes=<n>         exit after <n> retry passes (-1=infinity) [0]\n"
-               "  -R, --reverse                  reverse direction of some copy operations\n"
+               "  -R, --reverse                  reverse the direction of all passes\n"
                "  -s, --size=<bytes>             maximum size of input data to be copied\n"
                "  -S, --sparse                   use sparse writes for output file\n"
                "  -t, --truncate                 truncate output file to zero size\n"
                "  -T, --timeout=<interval>       maximum time since last successful read\n"
+               "  -u, --unidirectional           run all passes in the same direction\n"
                "  -v, --verbose                  be verbose (a 2nd -v gives more)\n"
                "  -w, --ignore-write-errors      make fill mode ignore write errors\n"
                "  -x, --extend-outfile=<bytes>   extend outfile size to be at least this long\n"
+               "  -X, --exit-on-error            exit after the first read error\n"
                "  -1, --log-rates=<file>         log rates and error sizes in file\n"
                "  -2, --log-reads=<file>         log all read operations in file\n"
+               "      --ask                      ask for confirmation before starting the copy\n"
+               "      --cpass=<n>[,<n>]          select what copying pass(es) to run\n"
+               "      --pause=<interval>         time to wait between passes [0]\n"
                "Numbers may be in decimal, hexadecimal or octal, and may be followed by a\n"
                "multiplier: s = sectors, k = 1000, Ki = 1024, M = 10^6, Mi = 2^20, etc...\n"
                "Time intervals have the format 1[.5][smhd] or 1/2[smhd].\n"
@@ -258,7 +264,7 @@ int do_fill( const long long offset, Domain & domain,
     { show_error( "Output file is not seekable." ); return 1; }
 
   if( verbosity >= 0 )
-    std::printf( "\n\n%s %s\n", Program_name, PROGVERSION );
+    std::printf( "%s %s\n", Program_name, PROGVERSION );
   if( verbosity >= 1 )
     {
     std::printf( "About to fill with data from %s blocks of %s marked %s\n",
@@ -310,7 +316,7 @@ int do_generate( const long long offset, Domain & domain,
     { show_error( "Output file is not seekable." ); return 1; }
 
   if( verbosity >= 0 )
-    std::printf( "\n\n%s %s\n", Program_name, PROGVERSION );
+    std::printf( "%s %s\n", Program_name, PROGVERSION );
   if( verbosity >= 1 )
     {
     std::printf( "About to generate an approximate logfile for %s and %s\n",
@@ -325,13 +331,66 @@ int do_generate( const long long offset, Domain & domain,
   }
 
 
+const char * device_id_or_size( const int fd )
+  {
+  static char buf[32];
+  const char * p = device_id( fd );
+  if( !p )
+    {
+    const long long size = lseek( fd, 0, SEEK_END );
+    snprintf( buf, sizeof buf, "%lld", size );
+    p = buf;
+    }
+  return p;
+  }
+
+const char * device_id_or_size( const char * const name )
+  {
+  const int fd = open( name, O_RDONLY );
+  const char * const p = ( fd >= 0 ) ? device_id_or_size( fd ) : "";
+  if( fd >= 0 ) close( fd );
+  return p;
+  }
+
+
+void about_to_copy( const Rescuebook & rescuebook, const char * const iname,
+                    const char * const oname, const int ides, const bool ask )
+  {
+  if( ask || verbosity >= 0 )
+    std::printf( "%s %s\n", Program_name, PROGVERSION );
+  if( ask || verbosity >= 1 )
+    {
+    std::string iid, oid;
+    if( ask || verbosity >= 2 )
+      {
+      iid = " ["; iid += device_id_or_size( ides ); iid += ']';
+      oid = " ["; oid += device_id_or_size( oname ); oid += ']';
+      }
+    std::printf( "About to copy %sBytes from %s%s to %s%s.\n",
+                 rescuebook.domain().full() ? "an unknown number of " :
+                   format_num( rescuebook.domain().in_size() ),
+                 iname, iid.c_str(), oname, oid.c_str() );
+    }
+  }
+
+
+bool user_agrees_ids( const Rescuebook & rescuebook, const char * const iname,
+                      const char * const oname, const int ides )
+  {
+  about_to_copy( rescuebook, iname, oname, ides, true );
+  std::printf( "Proceed (y/N)? " );
+  std::fflush( stdout );
+  return ( std::tolower( std::fgetc( stdin ) ) == 'y' );
+  }
+
+
 int do_rescue( const long long offset, Domain & domain,
                const Domain * const test_domain, const Rb_options & rb_opts,
                const char * const iname, const char * const oname,
                const char * const logname, const int cluster,
                const int hardbs, const int o_trunc,
-               const bool preallocate, const bool synchronous,
-               const bool verify_input_size )
+               const bool ask, const bool preallocate,
+               const bool synchronous, const bool verify_input_size )
   {
   const int ides = open( iname, O_RDONLY | rb_opts.o_direct | O_BINARY );
   if( ides < 0 )
@@ -376,6 +435,8 @@ int do_rescue( const long long offset, Domain & domain,
     }
   if( rescuebook.read_only() ) return not_writable( logname );
 
+  if( ask && !user_agrees_ids( rescuebook, iname, oname, ides ) ) return 1;
+
   const int odes = open( oname, O_CREAT | O_WRONLY | o_trunc | O_BINARY,
                          outmode );
   if( odes < 0 )
@@ -397,16 +458,9 @@ int do_rescue( const long long offset, Domain & domain,
   if( !read_logger.open_file() )
     { show_error( "Can't open file for logging reads", errno ); return 1; }
 
-  if( verbosity >= 0 )
-    std::printf( "\n\n%s %s\n", Program_name, PROGVERSION );
+  if( !ask ) about_to_copy( rescuebook, iname, oname, ides, false );
   if( verbosity >= 1 )
     {
-    if( rescuebook.domain().full() )
-      std::printf( "About to copy an unknown number of bytes from %s to %s\n",
-                   iname, oname );
-    else
-      std::printf( "About to copy %sBytes from %s to %s\n",
-                   format_num( rescuebook.domain().in_size() ), iname, oname );
     std::printf( "    Starting positions: infile = %sB,  outfile = %sB\n",
                  format_num( rescuebook.domain().pos() ),
                  format_num( rescuebook.domain().pos() + rescuebook.offset() ) );
@@ -432,30 +486,33 @@ int do_rescue( const long long offset, Domain & domain,
       if( rescuebook.max_retries >= 0 )
         { nl = true;
           std::printf( "Max retry passes: %d    ", rescuebook.max_retries ); }
-      if( nl ) { nl = false; std::printf( "\n" ); }
+      if( nl ) { nl = false; std::fputc( '\n', stdout ); }
 
       if( rescuebook.min_read_rate == 0 )
         { nl = true; std::printf( "Min read rate: auto    " ); }
       else if( rescuebook.min_read_rate > 0 )
         { nl = true; std::printf( "Min read rate:  %6sB/s    ",
                                   format_num( rescuebook.min_read_rate, 99999 ) ); }
+      if( rescuebook.pause > 0 )
+        { nl = true; std::printf( "Pause: %s    ",
+                                  format_time( rescuebook.pause ) ); }
       if( rescuebook.timeout >= 0 )
-        { nl = true; std::printf( "Max time since last successful read: %s",
+        { nl = true; std::printf( "Timeout: %s",
                                   format_time( rescuebook.timeout ) ); }
-      if( nl ) { nl = false; std::printf( "\n" ); }
+      if( nl ) { nl = false; std::fputc( '\n', stdout ); }
 
       std::printf( "Direct: %s    ", rescuebook.o_direct ? "yes" : "no" );
       std::printf( "Sparse: %s    ", rescuebook.sparse ? "yes" : "no" );
-      std::printf( "Split: %s    ", !rescuebook.nosplit ? "yes" : "no" );
       std::printf( "Trim: %s    ", !rescuebook.notrim ? "yes" : "no" );
+      std::printf( "Scrape: %s    ", !rescuebook.noscrape ? "yes" : "no" );
       std::printf( "Truncate: %s    ", o_trunc ? "yes" : "no" );
-      std::printf( "\n" );
+      std::fputc( '\n', stdout );
       if( rescuebook.complete_only )
         { nl = true; std::printf( "Complete only    " ); }
       if( rescuebook.reverse ) { nl = true; std::printf( "Reverse mode" ); }
-      if( nl ) { nl = false; std::printf( "\n" ); }
+      if( nl ) { nl = false; std::fputc( '\n', stdout ); }
       }
-    std::printf( "\n" );
+    std::fputc( '\n', stdout );
     }
   return rescuebook.do_rescue( ides, odes );
   }
@@ -468,6 +525,28 @@ int do_rescue( const long long offset, Domain & domain,
 
 namespace {
 
+void parse_cpass( const std::string & arg, Rb_options & rb_opts )
+  {
+  bool error = arg.empty();
+  rb_opts.cpass_bitset = 0;
+  for( unsigned i = 0; i < arg.size(); ++i )
+    {
+    const char ch = arg[i];
+    if( ch < '0' || ch > '3' ) { error = true; break; }
+    if( ch > '0' ) rb_opts.cpass_bitset |= ( 1 << ( ch - '1' ) );
+    if( i + 1 < arg.size() )
+      {
+      if( arg[i+1] == ',' && i + 2 < arg.size() ) ++i;
+      else { error = true; break; }
+      }
+    }
+  if( error )
+    {
+    show_error( "Bad list of passes in option '--cpass'." );
+    std::exit( 1 );
+    }
+  }
+
 void parse_skipbs( const char * const arg, Rb_options & rb_opts,
                    const int hardbs )
   {
@@ -508,6 +587,7 @@ bool Rescuebook::reopen_infile()
 
 int main( const int argc, const char * const argv[] )
   {
+  enum Optcode { opt_ask = 256, opt_cpa, opt_pau };
   long long ipos = 0;
   long long opos = -1;
   long long max_size = -1;
@@ -521,6 +601,7 @@ int main( const int argc, const char * const argv[] )
   int o_trunc = 0;
   Mode program_mode = m_none;
   Rb_options rb_opts;
+  bool ask = false;
   bool force = false;
   bool ignore_write_errors = false;
   bool loose = false;
@@ -532,7 +613,6 @@ int main( const int argc, const char * const argv[] )
   command_line = argv[0];
   for( int i = 1; i < argc; ++i )
     { command_line += ' '; command_line += argv[i]; }
-  initial_time_ = std::time( 0 );
 
   const Arg_parser::Option options[] =
     {
@@ -540,7 +620,6 @@ int main( const int argc, const char * const argv[] )
     { '2', "log-reads",           Arg_parser::yes },
     { 'a', "min-read-rate",       Arg_parser::yes },
     { 'A', "try-again",           Arg_parser::no  },
-    { 'b', "block-size",          Arg_parser::yes },
     { 'b', "sector-size",         Arg_parser::yes },
     { 'B', "binary-prefixes",     Arg_parser::no  },
     { 'c', "cluster-size",        Arg_parser::yes },
@@ -557,27 +636,32 @@ int main( const int argc, const char * const argv[] )
     { 'i', "input-position",      Arg_parser::yes },
     { 'I', "verify-input-size",   Arg_parser::no  },
     { 'K', "skip-size",           Arg_parser::yes },
-    { 'l', "logfile-size",        Arg_parser::yes },
     { 'L', "loose-domain",        Arg_parser::no  },
     { 'm', "domain-logfile",      Arg_parser::yes },
     { 'M', "retrim",              Arg_parser::no  },
+    { 'n', "no-scrape",           Arg_parser::no  },
     { 'n', "no-split",            Arg_parser::no  },
     { 'N', "no-trim",             Arg_parser::no  },
     { 'o', "output-position",     Arg_parser::yes },
     { 'O', "reopen-on-error",     Arg_parser::no  },
     { 'p', "preallocate",         Arg_parser::no  },
+    { 'P', "data-preview",        Arg_parser::maybe },
     { 'q', "quiet",               Arg_parser::no  },
     { 'r', "retry-passes",        Arg_parser::yes },
     { 'R', "reverse",             Arg_parser::no  },
     { 's', "size",                Arg_parser::yes },
-    { 's', "max-size",            Arg_parser::yes },
     { 'S', "sparse",              Arg_parser::no  },
     { 't', "truncate",            Arg_parser::no  },
     { 'T', "timeout",             Arg_parser::yes },
+    { 'u', "unidirectional",      Arg_parser::no  },
     { 'v', "verbose",             Arg_parser::no  },
     { 'V', "version",             Arg_parser::no  },
     { 'w', "ignore-write-errors", Arg_parser::no  },
     { 'x', "extend-outfile",      Arg_parser::yes },
+    { 'X', "exit-on-error",       Arg_parser::no  },
+    { opt_ask, "ask",             Arg_parser::no  },
+    { opt_cpa, "cpass",           Arg_parser::yes },
+    { opt_pau, "pause",           Arg_parser::yes },
     {  0 , 0,                     Arg_parser::no  } };
 
   const Arg_parser parser( argc, argv, options );
@@ -619,15 +703,16 @@ int main( const int argc, const char * const argv[] )
       case 'i': ipos = getnum( arg, hardbs, 0 ); break;
       case 'I': verify_input_size = true; break;
       case 'K': parse_skipbs( arg, rb_opts, hardbs ); break;
-      case 'l': rb_opts.max_logfile_size = getnum( arg, 0, 1, INT_MAX ); break;
       case 'L': loose = true; break;
       case 'm': set_name( &domain_logfile_name, arg, code ); break;
       case 'M': rb_opts.retrim = true; break;
-      case 'n': rb_opts.nosplit = true; break;
+      case 'n': rb_opts.noscrape = true; break;
       case 'N': rb_opts.notrim = true; break;
       case 'o': opos = getnum( arg, hardbs, 0 ); break;
       case 'O': rb_opts.reopen_on_error = true; break;
       case 'p': preallocate = true; break;
+      case 'P': rb_opts.preview_lines = arg[0] ? getnum( arg, 0, 1, 32 ) : 3;
+                break;
       case 'q': verbosity = -1; break;
       case 'r': rb_opts.max_retries = getnum( arg, 0, -1, INT_MAX ); break;
       case 'R': rb_opts.reverse = true; break;
@@ -635,10 +720,15 @@ int main( const int argc, const char * const argv[] )
       case 'S': rb_opts.sparse = true; break;
       case 't': o_trunc = O_TRUNC; break;
       case 'T': rb_opts.timeout = parse_time_interval( arg ); break;
+      case 'u': rb_opts.unidirectional = true; break;
       case 'v': if( verbosity < 4 ) ++verbosity; break;
       case 'V': show_version(); return 0;
       case 'w': ignore_write_errors = true; break;
       case 'x': rb_opts.min_outfile_size = getnum( arg, hardbs, 1 ); break;
+      case 'X': rb_opts.exit_on_error = true; break;
+      case opt_ask: ask = true; break;
+      case opt_cpa: parse_cpass( parser.argument( argind ), rb_opts ); break;
+      case opt_pau: rb_opts.pause = parse_time_interval( arg ); break;
       default : internal_error( "uncaught option." );
       }
     } // end process options
@@ -667,15 +757,21 @@ int main( const int argc, const char * const argv[] )
   switch( program_mode )
     {
     case m_fill:
+      if( ask )
+        { show_error( "Option '--ask' is incompatible with fill mode.", 0, true );
+          return 1; }
       if( rb_opts != Rb_options() || test_mode_logfile_name ||
           verify_input_size || preallocate || o_trunc )
-        show_error( "warning: Options -aACdeEHIKlMnOprRStTx are ignored in fill mode." );
+        show_error( "warning: Options -aACdeEHIKlMnOpPrRStTuxX are ignored in fill mode." );
       return do_fill( opos - ipos, domain, iname, oname, logname, cluster,
                       hardbs, filltypes, ignore_write_errors, synchronous );
     case m_generate:
+      if( ask )
+        { show_error( "Option '--ask' is incompatible with generate mode.", 0, true );
+          return 1; }
       if( rb_opts != Rb_options() || synchronous || test_mode_logfile_name ||
           verify_input_size || preallocate || o_trunc || ignore_write_errors )
-        show_error( "warning: Options -aACdDeEHIKlMnOprRStTwx are ignored in generate mode." );
+        show_error( "warning: Options -aACdDeEHIKlMnOpPrRStTuwxX are ignored in generate mode." );
       return do_generate( opos - ipos, domain, iname, oname, logname,
                           cluster, hardbs );
     case m_none:
@@ -686,7 +782,7 @@ int main( const int argc, const char * const argv[] )
       const Domain * const test_domain = test_mode_logfile_name ?
         new Domain( 0, -1, test_mode_logfile_name, loose ) : 0;
       int tmp = do_rescue( opos - ipos, domain, test_domain, rb_opts, iname,
-                           oname, logname, cluster, hardbs, o_trunc,
+                           oname, logname, cluster, hardbs, o_trunc, ask,
                            preallocate, synchronous, verify_input_size );
       if( test_domain ) delete test_domain;
       return tmp;
